{------------------------------------------------------------------------------
  pdScriptAPI.pas

  written by  Precision software & consulting
              e-mail: info@be-precision.com
              web: http://www.be-precision.com

  Purpose:    pdScript embedded interpreter API unit for Delphi developers

  The source code is given as is. The author is not responsible
  for any possible damage done due to the use of this code.
  This unit can be freely used in any application. The complete
  source code remains property of the author and may not be distributed,
  published, given or sold in any form as such. No parts of the source
  code can be included in any other component or application without
  written authorization of the author.

  Copyright  2008 - 2011  Precision software & consulting
  All rights reserved

------------------------------------------------------------------------------
  History:

  - Version: 1.9
    * added - Overloaded Execute_pdScriptFunction routine to execute simple procedures and functions (without parameters) from within previously executed scripts

  - Version: 1.7
    * added - Routines for executing the scripts with "keep-alive" flag
    * added - Routines for executing procedures and functions from within previously executed scripts
    * added - More detailed comments

  - Version: 1.6
    * first public release
------------------------------------------------------------------------------}

unit pdScriptAPI;

interface

uses
  Windows;

const
  pdScriptDllName = 'pdScriptE.dll';      // full featured interpreter
  pdScriptDllNameCL = 'pdScriptECL.dll';  // basic interpreter (no support for VCL components, GUI forms, etc.)

  // Predefined return values of the script
  PDSCRIPT_RESULT_OK = 0;       // Default return value of the script
  PDSCRIPT_RESULT_ERROR = 1;    // Predefined return value - error
  PDSCRIPT_RESULT_WARNING = 2;  // Predefined return value - warning
  PDSCRIPT_RESULT_HINT = 3;     // Predefined return value - hint
  // Return values for the compiler mode (SyntaxCheckOnly is True, or /PDSC switch is specified in Params):
  PDSCRIPT_COMPILE_OK = 0;      // Script was successfuly compiled (syntax check o.k.)
  PDSCRIPT_COMPILE_ERROR = 1;   // Script compile failed
  // Predefined return values (for the situations, when the script file cannot be executed):
  PDSCRIPT_NOT_SPECIFIED = -1;  // The script file has not been specified
  PDSCRIPT_NOT_FOUND = -2;      // The script file was not found
  // Return values for fatal errors:
  PDSCRIPT_FATAL_ERROR = -3;    // Unexpected exception (for both compiler mode and run-time):
  PDSCRIPT_NOT_LOADED = -4;     // pdScript library is not loaded

type
  TpdScriptHostCallback = function(cCommand:PChar; cParams:PChar; cBuffer:PChar; cBufSize:Integer):Integer; stdcall;
  TpdScriptHostCallbackA = function(cCommand:PAnsiChar; cParams:PAnsiChar; cBuffer:PAnsiChar; cBufSize:Integer):Integer; stdcall;

var
  pdScriptLibHandle: HMODULE;   // Currently used pdScript library handle

  {
   Following routines are exported from pdScript library.
   You can use them directly, or you can use Delphi-friendly routines (see bellow: Run_pdScript, Inline_pdScript, etc.).
  }
  pdScript: function(ScriptFile: PWideChar; Params: PWideChar; SyntaxCheckOnly: BOOL): Integer; stdcall;
  {$EXTERNALSYM pdScript}
  pdScriptEx: function(ScriptFile: PWideChar; Params: PWideChar; SyntaxCheckOnly: BOOL; out pStdOut: PWideChar): Integer; stdcall;
  {$EXTERNALSYM pdScriptEx}
  pdScriptExLib: function(ScriptFile: PWideChar; Params: PWideChar; SyntaxCheckOnly: BOOL; out pStdOut: PWideChar; LibraryPath: PWideChar): Integer; stdcall;
  {$EXTERNALSYM pdScriptExLib}
  pdScriptInline: function(ScriptCode: PWideChar; ScriptDFM: PWideChar; Params: PWideChar; SyntaxCheckOnly: BOOL): Integer; stdcall;
  {$EXTERNALSYM pdScriptInline}
  pdScriptInlineEx: function(ScriptCode: PWideChar; ScriptDFM: PWideChar; Params: PWideChar; SyntaxCheckOnly: BOOL; out pStdOut: PWideChar): Integer; stdcall;
  {$EXTERNALSYM pdScriptInlineEx}
  pdScriptInlineExLib: function(ScriptCode: PWideChar; ScriptDFM: PWideChar; Params: PWideChar; SyntaxCheckOnly: BOOL; out pStdOut: PWideChar; LibraryPath: PWideChar): Integer; stdcall;
  {$EXTERNALSYM pdScriptInlineExLib}

  pdsExecuteFunction: function(pScript: THandle; FncName: PWideChar; Params: PWideChar; PrmDelim:PWideChar): Integer; stdcall;
  {$EXTERNALSYM pdsExecuteFunction}
  pdsExecuteFunctionEx: function(pScript: THandle; FncName: PWideChar; Params: PWideChar; PrmDelim: PWideChar; VarParams: BOOL; out FncResult:PWideChar): Integer; stdcall;
  {$EXTERNALSYM pdsExecuteFunctionEx}

  pdsGetScriptForm: function(pScript: THandle):HWND; stdcall;
  {$EXTERNALSYM pdsGetScriptForm}
  pdsGetScriptResult: function(pScript: THandle): Integer; stdcall;
  {$EXTERNALSYM pdsGetScriptResult}
  pdsGetScriptResultEx: function(pScript: THandle; out pStdOut: PWideChar): Integer; stdcall;
  {$EXTERNALSYM pdsGetScriptResultEx}

  pdsFreeScript: procedure(pScript: THandle); stdcall;
  {$EXTERNALSYM pdsFreeScript}
  pdsFreeMem: procedure(p: Pointer); stdcall;
  {$EXTERNALSYM pdsFreeMem}

  // ANSI variants of above routines
  pdScriptA: function(ScriptFile: PAnsiChar; Params: PAnsiChar; SyntaxCheckOnly: BOOL): Integer; stdcall;
  {$EXTERNALSYM pdScriptA}
  pdScriptExA: function(ScriptFile: PAnsiChar; Params: PAnsiChar; SyntaxCheckOnly: BOOL; out pStdOut: PAnsiChar): Integer; stdcall;
  {$EXTERNALSYM pdScriptExA}
  pdScriptExLibA: function(ScriptFile: PAnsiChar; Params: PAnsiChar; SyntaxCheckOnly: BOOL; out pStdOut: PAnsiChar; LibraryPath: PAnsiChar): Integer; stdcall;
  {$EXTERNALSYM pdScriptExLibA}
  pdScriptInlineA: function(ScriptCode: PAnsiChar; ScriptDFM: PAnsiChar; Params: PAnsiChar; SyntaxCheckOnly: BOOL): Integer; stdcall;
  {$EXTERNALSYM pdScriptInlineA}
  pdScriptInlineExA: function(ScriptCode: PAnsiChar; ScriptDFM: PAnsiChar; Params: PAnsiChar; SyntaxCheckOnly: BOOL; out pStdOut: PAnsiChar): Integer; stdcall;
  {$EXTERNALSYM pdScriptInlineExA}
  pdScriptInlineExLibA: function(ScriptCode: PAnsiChar; ScriptDFM: PAnsiChar; Params: PAnsiChar; SyntaxCheckOnly: BOOL; out pStdOut: PAnsiChar; LibraryPath: PAnsiChar): Integer; stdcall;
  {$EXTERNALSYM pdScriptInlineExLibA}

  pdsExecuteFunctionA: function(pScript: THandle; FncName: PAnsiChar; Params: PAnsiChar; PrmDelim: PAnsiChar): Integer; stdcall;
  {$EXTERNALSYM pdsExecuteFunctionA}
  pdsExecuteFunctionExA: function(pScript: THandle; FncName: PAnsiChar; Params: PAnsiChar; PrmDelim: PAnsiChar; VarParams: BOOL; out FncResult: PAnsiChar): Integer; stdcall;
  {$EXTERNALSYM pdsExecuteFunctionExA}

  pdsGetScriptResultExA: function(pScript: THandle; out pStdOut: PAnsiChar): Integer; stdcall;
  {$EXTERNALSYM pdsGetScriptResultExA}


{
  Following are Delphi-friendly routines to execute and use scripts via pdScript library
}
// Load selected pdScript library
function pdScriptLoad(LibraryFileName: string = ''):HMODULE;
// Release currently used pdScript library
function pdScriptUnload: Boolean;
// Checks if the pdScript library is available. If not, then tries to load it.
function pdScriptReady: Boolean;

{ Basic routine to execute (or syntax-check) a desired script file. Returns compilation result (on unsuccessful compile) or script ExitCode. }
function Run_pdScript(const ScriptFile: string; const Params: string = ''; SyntaxCheckOnly: Boolean = False): Integer;
{ General routine to execute (or syntax-check) a desired script file. Returns compilation result (on unsuccessful compile)
  or script ExitCode and script output buffer. Allows you to use your own library path to supporting code snippets (.inc files). }
function Run_pdScriptEx(const ScriptFile, Params: string; SyntaxCheckOnly: Boolean; var pStdOut: string; const LibraryPath: string = ''): Integer;
{ Inline_pdScript is similar to Run_pdScript, but allows you to run an inline script, instead of script file. }
function Inline_pdScript(const ScriptCode, ScriptDFM: string; const Params: string = ''; SyntaxCheckOnly: Boolean = False): Integer;
{ Inline_pdScriptEx is similar to Run_pdScriptEx, but allows you to run an inline script, instead of script file. }
function Inline_pdScriptEx(const ScriptCode, ScriptDFM, Params: string; SyntaxCheckOnly: Boolean; var pStdOut: string; const LibraryPath: string = ''): Integer;

{ Start_pdScript runs the script with keep-alive flag (the form id displayed as non-modal, script remains in memory).
  AppHandle can be Application.Handle or Application.MainForm.Handle (ie. depending on which window is used as main in Windows shell).
  If you pass a ParentCtrl handle, the form is created as a child of this control.
  Return value is a handle of the script, that must be released by explicit call to Release_pdScript().
  If the script cannot be executed, return value is equal or lower than 1 (PDSCRIPT_RESULT_ERROR, etc. }
function Start_pdScript(const ScriptFile: string; Params: string = ''; AppHandle: HWND = 0; ParentCtrl: HWND = 0;
    Left: Integer = -1; Top: Integer = -1; Width: Integer = 0; Height: Integer = 0): Integer;
{ Start_pdScriptEx is the same routine as Start_pdScript, but returns an additional info about possible unsuccessful compile
  of the script (in pStdOut) and allows you to use your own library path to supporting code snippets (.inc files). }
function Start_pdScriptEx(const ScriptFile: string; Params: string; var pStdOut: string; const LibraryPath: string = '';
    AppHandle: HWND = 0; ParentCtrl: HWND = 0; Left: Integer = -1; Top: Integer = -1; Width: Integer = 0; Height: Integer = 0): Integer;
{ StartInline_pdScript is similar to Start_pdScript but allows you to run an inline script, instead of script file. }
function StartInline_pdScript(const ScriptCode, ScriptDFM: string; Params: string = ''; AppHandle: HWND = 0; ParentCtrl: HWND = 0;
    Left: Integer = -1; Top: Integer = -1; Width: Integer = 0; Height: Integer = 0): Integer;
{ StartInline_pdScriptEx is similar to Start_pdScriptEx but allows you to run an inline script, instead of script file. }
function StartInline_pdScriptEx(const ScriptCode, ScriptDFM: string; Params: string; var pStdOut: string; const LibraryPath: string = '';
    AppHandle: HWND = 0; ParentCtrl: HWND = 0; Left: Integer = -1; Top: Integer = -1; Width: Integer = 0; Height: Integer = 0): Integer;

// Releases previously executed pdScript instance (created by Start_pdScript, Start_pdScriptEx, StartInline_pdScript and StartInline_pdScriptEx routines. }
procedure Release_pdScript(pScript: THandle);

// Returns a main form handle of previously executed pdScript instance
function GetFormOf_pdScript(pScript: THandle):HWND;
// Returns current result of previously executed pdScript instance. If needed call it just before Release_pdScript.
function GetResultOf_pdScript(pScript: THandle): Integer;
// Returns current result and output buffer of previously executed pdScript instance. If needed call it just before Release_pdScript.
function GetResultOf_pdScriptEx(pScript: THandle; var pStdOut: string): Integer;

{ Executes a function or procedure that is declared inside the script code. Script must be previously executed with keep-alive flag (see Start_pdScript, StartInline_pdScript, etc.)
  By passing non-empty PrmDelim argument, you can tell the interpreter, that Params string contains multiple parameters delimited by that character (strictly).
  Returning value is 0 by default and -2 (empty script handle) or -1 (exception) on error. If the executed routine is a function with numeric or ordinal result,
  the returning value corresponds to that result. So you can use this implementation instead of more complex "Execute_pdScriptFunctionEx" variant. }
function Execute_pdScriptFunction(pScript: THandle; const FncName, Params: string; const PrmDelim: string = ''): Integer; overload;
// Similar as above, but returns string as a result (return value of function declared inside script, converted into string)
function Execute_pdScriptFunction(pScript: THandle; const FncName, Params: string; var FncResult: string; const PrmDelim: string = ''): Integer; overload;
// Similar as above, but can be used to call functions with "var" parameters.
function Execute_pdScriptFunctionEx(pScript: THandle; const FncName: string; var Params, FncResult: string; const PrmDelim: string = ''): Integer;
// Similar as above, but can be used to call procedures with "var" parameters.
function Execute_pdScriptProcedureEx(pScript: THandle; const ProcName: string; var Params: string; const PrmDelim: string = ''): Integer;
// Similar as above, but designed to call simple procedures and functions (without parameters)
function Execute_pdScriptFunction(pScript: THandle; const FncName:string): Integer; overload;

implementation

uses
  SysUtils;

function pdScriptLoad(LibraryFileName: string = ''): HMODULE;
begin
  if pdScriptLibHandle = 0 then
  begin
    if Length(LibraryFileName) = 0 then
      LibraryFileName := pdScriptDllName;
    pdScriptLibHandle := SafeLoadLibrary(PChar(LibraryFileName));
    if pdScriptLibHandle > 0 then
    begin
      @pdScript := GetProcAddress(pdScriptLibHandle, 'pdScript');
      @pdScriptEx := GetProcAddress(pdScriptLibHandle, 'pdScriptEx');
      @pdScriptExLib := GetProcAddress(pdScriptLibHandle, 'pdScriptExLib');
      @pdScriptInline := GetProcAddress(pdScriptLibHandle, 'pdScriptInline');
      @pdScriptInlineEx := GetProcAddress(pdScriptLibHandle, 'pdScriptInlineEx');
      @pdScriptInlineExLib := GetProcAddress(pdScriptLibHandle, 'pdScriptInlineExLib');
      @pdsExecuteFunction := GetProcAddress(pdScriptLibHandle, 'pdsExecuteFunction');
      @pdsExecuteFunctionEx := GetProcAddress(pdScriptLibHandle, 'pdsExecuteFunctionEx');
      @pdsGetScriptForm := GetProcAddress(pdScriptLibHandle, 'pdsGetScriptForm');
      @pdsGetScriptResult := GetProcAddress(pdScriptLibHandle, 'pdsGetScriptResult');
      @pdsGetScriptResultEx := GetProcAddress(pdScriptLibHandle, 'pdsGetScriptResultEx');
      @pdsFreeScript := GetProcAddress(pdScriptLibHandle, 'pdsFreeScript');
      @pdsFreeMem := GetProcAddress(pdScriptLibHandle, 'pdsFreeMem');

      @pdScriptA := GetProcAddress(pdScriptLibHandle, 'pdScriptA');
      @pdScriptExA := GetProcAddress(pdScriptLibHandle, 'pdScriptExA');
      @pdScriptExLibA := GetProcAddress(pdScriptLibHandle, 'pdScriptExLibA');
      @pdScriptInlineA := GetProcAddress(pdScriptLibHandle, 'pdScriptInlineA');
      @pdScriptInlineExA := GetProcAddress(pdScriptLibHandle, 'pdScriptInlineExA');
      @pdScriptInlineExLibA := GetProcAddress(pdScriptLibHandle, 'pdScriptInlineExLibA');
      @pdsExecuteFunctionA := GetProcAddress(pdScriptLibHandle, 'pdsExecuteFunctionA');
      @pdsExecuteFunctionExA := GetProcAddress(pdScriptLibHandle, 'pdsExecuteFunctionExA');
      @pdsGetScriptResultExA := GetProcAddress(pdScriptLibHandle, 'pdsGetScriptResultExA');
    end;
  end;
  Result := pdScriptLibHandle;
end;

function pdScriptUnload: Boolean;
begin
  if pdScriptLibHandle = 0 then
    Result := true
  else
  begin
    Result := FreeLibrary(pdScriptLibHandle);
    if Result then
    begin
      pdScriptLibHandle := 0;

      pdScript := nil;
      pdScriptEx := nil;
      pdScriptExLib := nil;
      pdScriptInline := nil;
      pdScriptInlineEx := nil;
      pdScriptInlineExLib := nil;

      pdScriptA := nil;
      pdScriptExA := nil;
      pdScriptExLibA := nil;
      pdScriptInlineA := nil;
      pdScriptInlineExA := nil;
      pdScriptInlineExLibA := nil;

      pdsGetScriptForm := nil;
      pdsGetScriptResult := nil;
      pdsGetScriptResultEx := nil;
      pdsGetScriptResultExA := nil;

      pdsFreeScript := nil;
      pdsFreeMem := nil;
    end;
  end;
end;

function pdScriptReady: Boolean;
begin
  if pdScriptLibHandle = 0 then
    Result := pdScriptLoad <> 0
  else
    Result := True;
end;

function Run_pdScript(const ScriptFile: string; const Params: string = ''; SyntaxCheckOnly: Boolean = False): Integer;
begin
  {$IFDEF UNICODE}
  if pdScriptReady and (@pdScript <> nil) then
    Result := pdScript(PWideChar(ScriptFile), PWideChar(Params), SyntaxCheckOnly)
  {$ELSE}
  if pdScriptReady and (@pdScriptA <> nil) then
    Result := pdScriptA(PAnsiChar(ScriptFile), PAnsiChar(Params), SyntaxCheckOnly)
  {$ENDIF}
  else
    Result := PDSCRIPT_NOT_LOADED;
end;

function Run_pdScriptEx(const ScriptFile, Params: string; SyntaxCheckOnly: Boolean; var pStdOut: string; const LibraryPath: string = ''): Integer;
{$IFDEF UNICODE}
var
  rep: PWideChar;
begin
  if pdScriptReady and (@pdScriptEx <> nil) and (@pdScriptExLib <> nil) and (@pdsFreeMem <> nil) then
  begin
    if Length(LibraryPath)=0 then
      Result:=pdScriptEx(PWideChar(ScriptFile), PWideChar(Params), SyntaxCheckOnly, rep)
    else
      Result:=pdScriptExLib(PWideChar(ScriptFile), PWideChar(Params), SyntaxCheckOnly, rep, PWideChar(LibraryPath));
{$ELSE}
var
  rep: PAnsiChar;
begin
  if pdScriptReady and (@pdScriptExA <> nil) and (@pdScriptExLibA <> nil) and (@pdsFreeMem <> nil) then
  begin
    if Length(LibraryPath)=0 then
      Result:=pdScriptExA(PAnsiChar(ScriptFile), PAnsiChar(Params), SyntaxCheckOnly, rep)
    else
      Result:=pdScriptExLibA(PAnsiChar(ScriptFile), PAnsiChar(Params), SyntaxCheckOnly, rep, PAnsiChar(LibraryPath));
{$ENDIF}
    pStdOut := rep;
    pdsFreeMem(rep);
  end
  else
    Result := PDSCRIPT_NOT_LOADED;
end;

function Inline_pdScript(const ScriptCode, ScriptDFM: string; const Params: string = ''; SyntaxCheckOnly: Boolean = False): Integer;
begin
  {$IFDEF UNICODE}
  if pdScriptReady and (@pdScriptInline <> nil) then
    Result:=pdScriptInline(PWideChar(ScriptCode), PWideChar(ScriptDFM), PWideChar(Params), SyntaxCheckOnly)
  {$ELSE}
  if pdScriptReady and (@pdScriptInlineA <> nil) then
    Result:=pdScriptInlineA(PAnsiChar(ScriptCode), PAnsiChar(ScriptDFM), PAnsiChar(Params), SyntaxCheckOnly)
  {$ENDIF}
  else
    Result := PDSCRIPT_NOT_LOADED;
end;

function Inline_pdScriptEx(const ScriptCode, ScriptDFM, Params: string; SyntaxCheckOnly: Boolean; var pStdOut: string; const LibraryPath: string = ''): Integer;
{$IFDEF UNICODE}
var
  rep: PWideChar;
begin
  if pdScriptReady and (@pdScriptInlineEx <> nil) and (@pdScriptInlineExLib <> nil) and (@pdsFreeMem <> nil) then
  begin
    if Length(LibraryPath)=0 then
      Result:=pdScriptInlineEx(PWideChar(ScriptCode), PWideChar(ScriptDFM), PWideChar(Params), SyntaxCheckOnly, rep)
    else
      Result:=pdScriptInlineExLib(PWideChar(ScriptCode), PWideChar(ScriptDFM), PWideChar(Params), SyntaxCheckOnly, rep, PWideChar(LibraryPath));
{$ELSE}
var
  rep: PAnsiChar;
begin
  if pdScriptReady and (@pdScriptInlineExA <> nil) and (@pdScriptInlineExLibA <> nil) and (@pdsFreeMem <> nil) then
  begin
    if Length(LibraryPath)=0 then
      Result:=pdScriptInlineExA(PAnsiChar(ScriptCode), PAnsiChar(ScriptDFM), PAnsiChar(Params), SyntaxCheckOnly, rep)
    else
      Result:=pdScriptInlineExLibA(PAnsiChar(ScriptCode), PAnsiChar(ScriptDFM), PAnsiChar(Params), SyntaxCheckOnly, rep, PAnsiChar(LibraryPath));
{$ENDIF}
    pStdOut := rep;
    pdsFreeMem(rep);
  end
  else
    Result := PDSCRIPT_NOT_LOADED;
end;

// Supporting routine to instantiate "/pds{<options>}" parameter for scripts executed with "keep-alive" flag.
procedure _UpdatePDSOptions(var Params:string; AppHandle: HWND = 0; ParentCtrl: HWND = 0;
    Left: Integer = -1; Top: Integer = -1; Width: Integer = 0; Height: Integer = 0);
var
  p:Integer;
  pdsOpt:string;
begin
  p:=Pos('/pds{',Params);
  if p=0 then
    pdsOpt:=''
  else
  begin
    pdsOpt:=Copy(Params,p+5,maxint);
    Delete(Params,p,maxint);
    p:=Pos('}',pdsOpt);
    if p>0 then
    begin
      if Copy(params,Length(params),1)='"' then
        Delete(params,Length(params),1);
      if Copy(pdsOpt,p+1,1)='"' then
        Params:=Params+Copy(pdsOpt,p+2,maxint)
      else
        Params:=Params+Copy(pdsOpt,p+1,maxint);
      Delete(pdsOpt,p,maxint);
    end;
  end;

  if Length(pdsOpt)>0 then
    pdsOpt:=pdsOpt+';';
  pdsOpt:=pdsOpt+'keep-alive=true';
  if AppHandle>0 then
    pdsOpt:=pdsOpt+';host-app='+IntToStr(AppHandle);
  if ParentCtrl>0 then
    pdsOpt:=pdsOpt+';host-parent='+IntToStr(ParentCtrl);
  if (Left>=0) or (Top>=0) or (Width>0) or (Height>0) then
    pdsOpt:=pdsOpt+';host-bounds='+Format('%d,%d,%d,%d',[Left,Top,Width,Height]);
  Params:=Params+' "/pds{'+pdsOpt+'}"';
end;

function Start_pdScript(const ScriptFile: string; Params: string = ''; AppHandle: HWND = 0; ParentCtrl: HWND = 0;
  Left: Integer = -1; Top: Integer = -1; Width: Integer = 0; Height: Integer = 0): Integer;
begin
  {$IFDEF UNICODE}
  if pdScriptReady and (@pdScript <> nil) and (@pdsFreeScript <> nil) then
  {$ELSE}
  if pdScriptReady and (@pdScriptA <> nil) and (@pdsFreeScript <> nil) then
  {$ENDIF}
  begin
    _UpdatePDSOptions(Params, AppHandle, ParentCtrl, Left, Top, Width, Height);
    {$IFDEF UNICODE}
    Result := pdScript(PWideChar(ScriptFile), PWideChar(Params), False)
    {$ELSE}
    Result := pdScriptA(PAnsiChar(ScriptFile), PAnsiChar(Params), False)
    {$ENDIF}
  end
  else
    Result := PDSCRIPT_NOT_LOADED;
end;

function Start_pdScriptEx(const ScriptFile: string; Params: string; var pStdOut: string; const LibraryPath: string = '';
    AppHandle: HWND = 0; ParentCtrl: HWND = 0; Left: Integer = -1; Top: Integer = -1; Width: Integer = 0; Height: Integer = 0): Integer;
var
  rep: {$IFDEF UNICODE} PWideChar {$ELSE} PAnsiChar {$ENDIF} ;
begin
  {$IFDEF UNICODE}
  if pdScriptReady and (@pdScriptEx <> nil) and (@pdScriptExLib <> nil) and (@pdsFreeScript <> nil) and (@pdsFreeMem <> nil) then
  {$ELSE}
  if pdScriptReady and (@pdScriptExA <> nil) and (@pdScriptExLibA <> nil) and (@pdsFreeScript <> nil) and (@pdsFreeMem <> nil) then
  {$ENDIF}
  begin
    _UpdatePDSOptions(Params, AppHandle, ParentCtrl, Left, Top, Width, Height);
    {$IFDEF UNICODE}
    if Length(LibraryPath)=0 then
      Result:=pdScriptEx(PWideChar(ScriptFile), PWideChar(Params), False, rep)
    else
      Result:=pdScriptExLib(PWideChar(ScriptFile), PWideChar(Params), False, rep, PWideChar(LibraryPath));
    {$ELSE}
    if Length(LibraryPath)=0 then
      Result:=pdScriptExA(PAnsiChar(ScriptFile), PAnsiChar(Params), False, rep)
    else
      Result:=pdScriptExLibA(PAnsiChar(ScriptFile), PAnsiChar(Params), False, rep, PAnsiChar(LibraryPath));
    {$ENDIF}
    pStdOut := rep;
    pdsFreeMem(rep);
  end
  else
    Result := PDSCRIPT_NOT_LOADED;
end;

function StartInline_pdScript(const ScriptCode, ScriptDFM: string; Params: string = ''; AppHandle: HWND = 0; ParentCtrl: HWND = 0;
    Left: Integer = -1; Top: Integer = -1; Width: Integer = 0; Height: Integer = 0): Integer;
begin
  {$IFDEF UNICODE}
  if pdScriptReady and (@pdScriptInline <> nil) and (@pdsFreeScript <> nil) then
  {$ELSE}
  if pdScriptReady and (@pdScriptInlineA <> nil) and (@pdsFreeScript <> nil) then
  {$ENDIF}
  begin
    _UpdatePDSOptions(Params, AppHandle, ParentCtrl, Left, Top, Width, Height);
    {$IFDEF UNICODE}
    Result:=pdScriptInline(PWideChar(ScriptCode), PWideChar(ScriptDFM), PWideChar(Params), False)
    {$ELSE}
    Result:=pdScriptInlineA(PAnsiChar(ScriptCode), PAnsiChar(ScriptDFM), PAnsiChar(Params), False)
    {$ENDIF}
  end
  else
    Result := PDSCRIPT_NOT_LOADED;
end;

function StartInline_pdScriptEx(const ScriptCode, ScriptDFM: string; Params: string; var pStdOut: string; const LibraryPath: string = '';
    AppHandle: HWND = 0; ParentCtrl: HWND = 0; Left: Integer = -1; Top: Integer = -1; Width: Integer = 0; Height: Integer = 0): Integer;
var
  rep: {$IFDEF UNICODE} PWideChar {$ELSE} PAnsiChar {$ENDIF} ;
begin
  {$IFDEF UNICODE}
  if pdScriptReady and (@pdScriptInlineEx <> nil) and (@pdScriptInlineExLib <> nil) and (@pdsFreeScript <> nil) and (@pdsFreeMem <> nil) then
  {$ELSE}
  if pdScriptReady and (@pdScriptInlineExLibA <> nil) and (@pdScriptInlineExLibA <> nil) and (@pdsFreeScript <> nil) and (@pdsFreeMem <> nil) then
  {$ENDIF}
  begin
    _UpdatePDSOptions(Params, AppHandle, ParentCtrl, Left, Top, Width, Height);
    {$IFDEF UNICODE}
    if Length(LibraryPath)=0 then
      Result:=pdScriptInlineEx(PWideChar(ScriptCode), PWideChar(ScriptDFM), PWideChar(Params), False, rep)
    else
      Result:=pdScriptInlineExLib(PWideChar(ScriptCode), PWideChar(ScriptDFM), PWideChar(Params), False, rep, PWideChar(LibraryPath));
    {$ELSE}
    if Length(LibraryPath)=0 then
      Result:=pdScriptInlineExA(PAnsiChar(ScriptCode), PAnsiChar(ScriptDFM), PAnsiChar(Params), False, rep)
    else
      Result:=pdScriptInlineExLibA(PAnsiChar(ScriptCode), PAnsiChar(ScriptDFM), PAnsiChar(Params), False, rep, PAnsiChar(LibraryPath));
    {$ENDIF}
    pStdOut := rep;
    pdsFreeMem(rep);
  end
  else
    Result := PDSCRIPT_NOT_LOADED;
end;

procedure Release_pdScript(pScript: THandle);
begin
  if (pdScriptLibHandle <> 0) and (@pdsFreeScript <> nil) then
    pdsFreeScript(pScript)
end;

function GetFormOf_pdScript(pScript: THandle):HWND;
begin
  if (pdScriptLibHandle <> 0) and (@pdsGetScriptForm <> nil) then
    Result:=pdsGetScriptForm(pScript)
  else
    Result:=0;
end;

function GetResultOf_pdScript(pScript: THandle): Integer;
begin
  if (pdScriptLibHandle <> 0) and (@pdsGetScriptResult <> nil) then
    Result:=pdsGetScriptResult(pScript)
  else
    Result:=PDSCRIPT_NOT_LOADED;
end;

function GetResultOf_pdScriptEx(pScript: THandle; var pStdOut: string): Integer;
{$IFDEF UNICODE}
var
  rep:  PWideChar;
begin
  if (pdScriptLibHandle <> 0) and (@pdsGetScriptResultEx <> nil) then
  begin
    Result:=pdsGetScriptResultEx(pScript, rep);
{$ELSE}
var
  rep:  PAnsiChar;
begin
  if (pdScriptLibHandle <> 0) and (@pdsGetScriptResultExA <> nil) then
  begin
    Result:=pdsGetScriptResultExA(pScript, rep);
{$ENDIF}
    pStdOut := rep;
    pdsFreeMem(rep);
  end
  else
    Result := PDSCRIPT_NOT_LOADED;
end;

function Execute_pdScriptFunction(pScript: THandle; const FncName:string): Integer;
begin
  if (pdScriptLibHandle <> 0) and (@pdsExecuteFunction <> nil) then
  begin
    {$IFDEF UNICODE}
    Result:=pdsExecuteFunction(pScript, PWideChar(FncName), nil, nil)
    {$ELSE}
    Result:=pdsExecuteFunctionA(pScript, PAnsiChar(FncName), nil, nil)
    {$ENDIF}
  end
  else
    Result:=PDSCRIPT_NOT_LOADED;
end;

function Execute_pdScriptFunction(pScript: THandle; const FncName, Params: string; const PrmDelim: string = ''): Integer;
begin
  if (pdScriptLibHandle <> 0) and (@pdsExecuteFunction <> nil) then
  begin
    {$IFDEF UNICODE}
    if PrmDelim='' then
      Result:=pdsExecuteFunction(pScript, PWideChar(FncName), PWideChar(Params), nil)
    else
      Result:=pdsExecuteFunction(pScript, PWideChar(FncName), PWideChar(Params), PWideChar(PrmDelim))
    {$ELSE}
    if PrmDelim='' then
      Result:=pdsExecuteFunctionA(pScript, PAnsiChar(FncName), PAnsiChar(Params), nil)
    else
      Result:=pdsExecuteFunctionA(pScript, PAnsiChar(FncName), PAnsiChar(Params), PAnsiChar(PrmDelim))
    {$ENDIF}
  end
  else
    Result:=PDSCRIPT_NOT_LOADED;
end;

function Execute_pdScriptFunction(pScript: THandle; const FncName, Params: string; var FncResult: string; const PrmDelim: string = ''): Integer;
var
  rep: {$IFDEF UNICODE} PWideChar {$ELSE} PAnsiChar {$ENDIF} ;
begin
  if (pdScriptLibHandle <> 0) and (@pdsExecuteFunctionEx <> nil) then
  begin
    {$IFDEF UNICODE}
    if PrmDelim='' then
      Result:=pdsExecuteFunctionEx(pScript, PWideChar(FncName), PWideChar(Params), nil, False, rep)
    else
      Result:=pdsExecuteFunctionEx(pScript, PWideChar(FncName), PWideChar(Params), PWideChar(PrmDelim), False, rep);
    {$ELSE}
    if PrmDelim='' then
      Result:=pdsExecuteFunctionExA(pScript, PAnsiChar(FncName), PAnsiChar(Params), nil, False, rep)
    else
      Result:=pdsExecuteFunctionExA(pScript, PAnsiChar(FncName), PAnsiChar(Params), PAnsiChar(PrmDelim), False, rep);
    {$ENDIF}
    case Result of
      -1:FncResult:=rep; // Error message
      -2:FncResult:='';
    else
      FncResult := rep;
    end;
    pdsFreeMem(rep);
  end
  else
    Result:=PDSCRIPT_NOT_LOADED;
end;

function Execute_pdScriptFunctionEx(pScript: THandle; const FncName: string; var Params, FncResult: string; const PrmDelim: string = ''): Integer;
var
  p: Integer;
  rep: {$IFDEF UNICODE} PWideChar {$ELSE} PAnsiChar {$ENDIF} ;
begin
  if (pdScriptLibHandle <> 0) and (@pdsExecuteFunctionEx <> nil) then
  begin
    {$IFDEF UNICODE}
    if PrmDelim='' then
      Result:=pdsExecuteFunctionEx(pScript, PWideChar(FncName), PWideChar(Params), nil, True, rep)
    else
      Result:=pdsExecuteFunctionEx(pScript, PWideChar(FncName), PWideChar(Params), PWideChar(PrmDelim), True, rep);
    {$ELSE}
    if PrmDelim='' then
      Result:=pdsExecuteFunctionExA(pScript, PAnsiChar(FncName), PAnsiChar(Params), nil, True, rep)
    else
      Result:=pdsExecuteFunctionExA(pScript, PAnsiChar(FncName), PAnsiChar(Params), PAnsiChar(PrmDelim), True, rep);
    {$ENDIF}
    case Result of
      -1:FncResult:=rep; // Error message
      -2:FncResult:='';
    else
      begin
        if PrmDelim='' then
          FncResult:=rep
        else
        begin
          p:=LastDelimiter(PrmDelim,rep);
          if p=0 then
            FncResult:=rep
          else
          begin
            FncResult:=Copy(rep,p+1,maxint);
            Params:=Copy(rep,1,p-1);
          end;
        end
      end;
    end;
    pdsFreeMem(rep);
  end
  else
    Result:=PDSCRIPT_NOT_LOADED;
end;

function Execute_pdScriptProcedureEx(pScript: THandle; const ProcName: string; var Params: string; const PrmDelim: string = ''): Integer;
var
  f:string;
begin
  Result:=Execute_pdScriptFunctionEx(pScript, ProcName, Params, f, PrmDelim);
end;

initialization

finalization

  pdScriptUnload;

end.